# -*- coding: utf-8 -*-
"""Module for CRL manager
"""

# import basic stuff
import logging

# import own stuff
from tlsmate import tls
from tlsmate import cert_utils

# import other stuff
import requests
from cryptography import x509


class CrlManager(object):
    """Handles all CRL related operations and acts as a cache as well
    """

    def __init__(self, tlsmate=None):
        self._crls = {}
        self._recorder = tlsmate.recorder

    def add_crl(self, url, der_crl=None, pem_crl=None):
        """Adds a URL and the CRL to the cache.

        Either der_crl or pem_crl must be given.

        Arguments:
            url (str): the URL of the CRL
            der_crl(bytes): the CRL in DER format given as bytes
            pem_crl(bytes): the CRL in PEM format given as bytes
        """
        crl = None
        if der_crl is not None:
            crl = x509.load_der_x509_crl(der_crl)

        elif pem_crl is not None:
            crl = x509.load_pem_x509_crl(pem_crl)

        self._crls[url] = crl

    def _get_crl_obj(self, url):
        """Get the plain CRL object for a given URL.
        """

        if url not in self._crls:
            bin_crl = None
            self._recorder.trace(crl_url=url)
            try:
                if self._recorder.is_injecting():
                    bin_crl = self._recorder.inject(crl=None)

                else:
                    crl_resp = requests.get(url, timeout=5)
                    if crl_resp.ok:
                        bin_crl = crl_resp.content

                    self._recorder.trace(crl=bin_crl)

            except Exception:
                self._crls[url] = None

            else:
                self.add_crl(url, der_crl=bin_crl)

        return self._crls[url]

    def get_crl_status(self, urls, serial_nbr, issuer, issuer_cert, timestamp):
        """Determines the CRL revocation status for a given cert/urls.

        Downloads the CRL (if a download fails, the next url is tried), if not yet
        cached, validates the CRL against the issuer & its signature and checks if
        the certificate is present in the CRL or not.

        Arguments:
            urls (list of str): a list of CRL-urls
            serial_nbr (int): the serial number of the certificate to check
            issuer (:obj:`x509.Name`): the issuer name of the cert to check
            issuer_cert (:obj:`tlsmate.cert.Certificate`): the certificate of the issuer

        Returns:
            :obj:`tlsmate.tls.CertCrlStatus`: the final status.
        """
        status = None
        for url in urls:
            logging.debug(f"downloading CRL from {url}")
            crl = self._get_crl_obj(url)
            if crl is None:
                status = tls.CertCrlStatus.CRL_DOWNLOAD_FAILED
                continue

            if not cert_utils.equal_names(issuer, issuer_cert.parsed.subject):
                return tls.CertCrlStatus.WRONG_CRL_ISSUER

            if crl.last_update > timestamp or crl.next_update < timestamp:
                return tls.CertCrlStatus.INVALID_TIMESTAMP

            if not crl.is_signature_valid(issuer_cert.parsed.public_key()):
                return tls.CertCrlStatus.CRL_SIGNATURE_INVALID

            if crl.get_revoked_certificate_by_serial_number(serial_nbr) is None:
                return tls.CertCrlStatus.NOT_REVOKED

            else:
                return tls.CertCrlStatus.REVOKED

        return status



# -*- coding: utf-8 -*-
"""Module containing the class for the client
"""
# import basic stuff
import time
from dataclasses import dataclass, field
from typing import List

# import own stuff
from tlsmate import tls
from tlsmate import ext
from tlsmate import structs
from tlsmate import utils
from tlsmate.msg import ClientHello
from tlsmate.connection import TlsConnection

# import other stuff


@dataclass
class ClientProfile(object):
    """Data class defining the TLS profile of the client.

    The client profile defines the protocol parameters, extensions and features
    which are supported by the client. The data is used when ``tlsmate`` generates
    the ClientHello message autonomously, e.g., when using::

        conn.send(msg.ClientHello)

        or

        conn.handshake()

    The client profile is used as well to check a received ServerHello for
    consistency, e.g., the version provided by the server must be in the
    version list of the client profile.

    Attributes:
        versions (list (:obj:`tlsmate.tls.Version` or int)): The list of protocol
            versions the client shall support. Note, that the highest version
            will be picked to offer it in a ClientHello. For the extension
            SupportedVersions this list will be ordered, so that the highest
            version comes first. Arbitrary integer values can be provided in the
            list as well, allowing to check if the server ignores unknown values.
            Default: []
        compression_methods (list (:obj:`tlsmate.tls.CompressionMethod`)):
            A list of supported compression methods. This list will be used to
            populate the compression list in the ClientHello message.
            Default: [:obj:`tlsmate.tls.CompressionMethod.NULL`]
        cipher_suites (list (:obj:`tlsmate.tls.CipherSuite` or int)):
            A list of cipher suites which will be offered to the server in the
            sequence given. Note, that arbitrary integer values are supported
            as well, allowing to check if the server ignores unknown values.
            Default: []
        support_sni (bool):
            An indication if the SNI extension shall be supported. If True, the SNI
            will be taken from the CLI parameter ``--sni`` or (if not given), from
            the host name. Default: True
        ec_point_formats (list (:obj:`tlsmate.tls.EcPointFormat` or int)): The list
            of ec-point formats supported by the client. If set to None, the
            extension will not be generated. Default: None
        supported_groups (list (:obj:`tlsmate.tls.SupportedGroups` or int)): The list
            of named groups supported by the client. If set to None, the
            extension will not be present in the ClientHello message. Note,
            that arbitrary integer values are supported as well, allowing to
            check if the server ignores unknown values. Default: None
        signature_algorithms (list (:obj:`tlsmate.tls.SignatureScheme` or int)):
            The list of signature algorithms supported by the client. If
            set to None, the extension will not be present in the ClientHello
            message. Note, that arbitrary integer values are supported as
            well, allowing to check if the server ignores unknown values.
            Default: None
        heartbeat_mode (:class:`tlsmate.tls.HeartbeatMode` or None): The mode which is
            offered in the heartbeat extension. If set to None, the extension
            will not be setup when using the :meth:`Client.client_hello`
            method. Default: None
        support_session_id (bool): An indication if the client shall support resumption
            via the session id. Received session ids from the server will be offered
            in subsequent handshakes. Default: False
        support_session_ticket (bool): An indication if the client shall
            support resumption via the extensions SessionTicket. Received
            session tickets from the server will be offered in subsequent
            handshakes. Default: False
        support_extended_master_secret (bool): An indication if the client shall support
            the extensions ExtendedMasterSecret. Default: False
        support_encrypt_then_mac (bool): An indication if the client shall support the
            EncryptThenMac extension. Default: False
        support_secure_renegotiation (bool): An indication if the client shall support
            secure renegotiation. This will generate the RenegotiationInfo extension.
            Default: False
        support_scsv_renegotiation (bool): An indication, if the cipher suite value
            TLS_EMPTY_RENEGOTIATION_INFO_SCSV shall be added to the cipher
            suite list. Only applicable if support_secure_renegotiation is
            True. Default: False
        support_psk (bool): An indication whether the client offers a PSK with
            the ClientHello (i.e. NewSessionTicket message have been received
            before). Default: False
        support_status_request (bool): An indication, if the extensions status request
            shall be supported. Default: False
        support_status_request_v2 (:obj:`tlsmate.tls.StatusType`): The status type of
            the request. NONE is used to suppress the extension.
            Default: NONE
        key_shares (list (:obj:`tlsmate.tls.SupportedGroups` or int)): The list
            of key shares supported for TLS1.3. Note, that arbitrary integer
            values are supported as well, allowing to check if the server
            ignores unknown values. Default: None
        psk_key_exchange_modes (list (:obj:`tlsmate.tls.PskKeyExchangeMode` or int)):
            The list of PSK key exchange modes used in the extension
            psk_key_exchange_modes. Note, that arbitrary integer values are
            supported as well, allowing to check if the server ignores unknown
            values. Default: None
        early_data (bytes): The application data to be sent with 0-RTT. TLS1.3 only.
            If None, then no early data will be sent. Early data can only be sent
            in subsequent abbreviated handshakes. Default: None
    """

    # common for all versions
    versions: List = field(default_factory=lambda: [])
    compression_methods: List = field(
        default_factory=lambda: [tls.CompressionMethod.NULL]
    )
    cipher_suites: List = field(default_factory=lambda: [])

    support_sni: bool = True
    ec_point_formats: List = None
    supported_groups: List = None
    signature_algorithms: List = None
    heartbeat_mode: tls.HeartbeatMode = None
    support_status_request: bool = False
    support_status_request_v2: tls.StatusType = tls.StatusType.NONE

    # TLS1.2 and below
    support_session_id: bool = False
    support_session_ticket: bool = False
    support_extended_master_secret: bool = False
    support_encrypt_then_mac: bool = False
    support_secure_renegotiation: bool = False
    support_scsv_renegotiation: bool = False

    # TLS1.3 specific
    support_psk: bool = False
    key_shares: List = None
    psk_key_exchange_modes: List = None
    early_data: bytes = None


class Client(object):
    """The class representing a TLS client

    Connections are initiated by the TLS client, and it also stores data across
    several connections, like session tickets.

    A TLS client has a dedicated client profile (:obj:`ClientProfile`) which
    controls the TLS connections, e.g. the supported TLS versions, the
    supported cipher suites and other attributes are defined there (e.g. set of
    supported groups, whether encrypt-then-mac is supported, etc.)

    The client profile is used for setting up TLS messages in case only the
    message class and not a message instance is provided in the test case.

    Attributes:
        config (:obj:`tlsmate.config.Configuration`): the configuration object
        session_state_id (:obj:`tlsmate.structs.SessionStateId`): the stored sessions
            state usable to resume a session with the session_id
        session_state_ticket (:obj:`tlsmate.structs.SessionStateTicket`): the stored
            sessions state usable to resume a session with the session_ticket extension
        psks (list of :obj:`tlsmate.structs.Psk`): The TLS1.3 PSKs
        alert_on_invalid_cert (bool): Controls the behavior in case a certificate or
            the complete certificate chain cannot successfully be validated.
            If True, the connection will be closed with a fatal alert. If False,
            the connection continues. The latter is useful for scanning a server, as
            the handshake would be aborted prematurely otherwise.
        server_issues (list of :obj:`tlsmate.tls.ServerIssue`): a list of severe server
            issues.
    """

    def __init__(self, tlsmate):
        """Initialize the client object

        Args:
            tlsmate (:obj:`tlsmate.tlsmate.TlsMate`): the tlsmate application object.
        """
        self._tlsmate = tlsmate
        self.config = tlsmate.config
        self._set_profile_modern()
        self.alert_on_invalid_cert = True
        self.session_state_ticket = None
        self.session_state_id = None
        self.psks = []
        self._host = None
        self.server_issues = []

    def report_server_issue(self, issue, message=None, extension=None):
        """Store a server issue, if not done

        Arguments:
            issue (:obj:tlsmate.tls.`ServerMalfunction`): the reason for the exception
            message (:obj:`tlsmate.tls.HandshakeType`): the message, if applicable
            extension (:obj:`tlsmate.tls.Extension`): the extension, if applicable
        """

        malfunction = structs.Malfunction(
            issue=issue, message=message, extension=extension
        )

        if malfunction not in self.server_issues:
            self.server_issues.append(malfunction)

    def init_profile(self, profile_values=None):
        """Resets the client profile to a very basic state

        :note:
            At least the versions and the cipher_suite list must be provided before
            this profile can be used.

        Compression methods is set to [:obj:`tlsmate.tls.CompressionMethod.NULL`], and
        by default the sni extension is enabled. Everything else is empty or disabled.

        Arguments:
            profile_values (:obj:`tlsmate.structs.ProfileValues`): the profile
                values to additionally use to initialize the client profile
        """

        self.profile = ClientProfile()
        if profile_values is not None:
            self.profile.versions = profile_values.versions[:]
            self.profile.cipher_suites = profile_values.cipher_suites[:]
            if profile_values.supported_groups:
                self.profile.supported_groups = profile_values.supported_groups[:]

            if profile_values.signature_algorithms:
                self.profile.signature_algorithms = profile_values.signature_algorithms[
                    :
                ]

            if profile_values.key_shares:
                self.profile.key_shares = profile_values.key_shares[:]

    def _set_profile_interoperability(self):
        """Define profile for interoperability, like used in modern browsers

        :note:
            This method is deprecated. Use :meth:`Client.set_profile` instead.

        Profile properties:
          - TLS Versions 1.0 - 1.3
          - ECDHE cipher & RSA-based key transport
          - AESGCM, AES, CHACHA_POLY and 3DES as last resort
          - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
          - Resumption, encrypt-then-mac, extended-master-secret
          - pskmode psk_dhe
        """
        self.profile = ClientProfile(
            versions=[
                tls.Version.TLS10,
                tls.Version.TLS11,
                tls.Version.TLS12,
                tls.Version.TLS13,
            ],
            cipher_suites=[
                tls.CipherSuite.TLS_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_RSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_3DES_EDE_CBC_SHA,
            ],
            supported_groups=[
                tls.SupportedGroups.X25519,
                tls.SupportedGroups.SECP256R1,
                tls.SupportedGroups.SECP384R1,
                tls.SupportedGroups.SECP521R1,
            ],
            signature_algorithms=[
                tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PKCS1_SHA256,
                tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PKCS1_SHA384,
                tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
                tls.SignatureScheme.RSA_PKCS1_SHA512,
                tls.SignatureScheme.ECDSA_SHA1,
                tls.SignatureScheme.RSA_PKCS1_SHA1,
            ],
            key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
            psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
            ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
            support_extended_master_secret=True,
            support_session_id=True,
            support_session_ticket=True,
            support_encrypt_then_mac=True,
        )

    def _set_profile_legacy(self):
        """Define profile for legacy like client

        :note:
            This method is deprecated. Use :meth:`Client.set_profile` instead.

        Profile properties:
          - TLS Versions 1.0 - 1.2
          - ECDHE cipher & DHE & RSA-based key transport
          - AESGCM, AES, CHACHA_POLY and 3DES as last resort
          - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
          - Resumption, encrypt-then-mac, extended-master-secret
        """

        self.profile = ClientProfile(
            versions=[tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12],
            cipher_suites=[
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_CBC_SHA256,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_CBC_SHA256,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_RSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA,
                tls.CipherSuite.TLS_RSA_WITH_3DES_EDE_CBC_SHA,
            ],
            supported_groups=[
                tls.SupportedGroups.X25519,
                tls.SupportedGroups.SECP256R1,
                tls.SupportedGroups.SECP384R1,
                tls.SupportedGroups.SECP521R1,
            ],
            signature_algorithms=[
                tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PKCS1_SHA256,
                tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PKCS1_SHA384,
                tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
                tls.SignatureScheme.RSA_PKCS1_SHA512,
                tls.SignatureScheme.ECDSA_SHA1,
                tls.SignatureScheme.RSA_PKCS1_SHA1,
            ],
            ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
            support_extended_master_secret=True,
            support_session_id=True,
            support_session_ticket=True,
            support_encrypt_then_mac=True,
        )

    def _set_profile_modern(self):
        """Define profile for "modern" configurations

        :note:
            This method is deprecated. Use :meth:`Client.set_profile` instead.

        Profile properties:
          - TLS Versions 1.2 + 1.3
          - ECDHE cipher
          - AESGCM, CHACHA_POLY
          - signatures: ECDSA + PSS_RSAE
          - Resumption, encrypt-then-mac, extended-master-secret
          - pskmode psk_dhe
        """

        self.profile = ClientProfile(
            versions=[tls.Version.TLS12, tls.Version.TLS13],
            cipher_suites=[
                tls.CipherSuite.TLS_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
                tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
            ],
            supported_groups=[
                tls.SupportedGroups.X25519,
                tls.SupportedGroups.SECP256R1,
                tls.SupportedGroups.SECP384R1,
                tls.SupportedGroups.SECP521R1,
            ],
            signature_algorithms=[
                tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            ],
            key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
            psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
            ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
            support_sni=True,
            support_extended_master_secret=True,
            support_session_id=True,
            support_session_ticket=True,
        )

    def _set_profile_tls13(self):
        """Define profile for TLS1.3 only.

        :note:
            This method is deprecated. Use :meth:`Client.set_profile` instead.

        Profile properties:
          - TLS Version 1.3
          - AESGCM + CHACHA_POLY
          - pskmode psk_dhe
        """

        self.profile = ClientProfile(
            versions=[tls.Version.TLS13],
            cipher_suites=[
                tls.CipherSuite.TLS_AES_128_GCM_SHA256,
                tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
                tls.CipherSuite.TLS_AES_256_GCM_SHA384,
            ],
            supported_groups=[
                tls.SupportedGroups.X25519,
                tls.SupportedGroups.SECP256R1,
                tls.SupportedGroups.SECP384R1,
                tls.SupportedGroups.SECP521R1,
            ],
            signature_algorithms=[
                tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PKCS1_SHA256,
                tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PKCS1_SHA384,
                tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
                tls.SignatureScheme.RSA_PKCS1_SHA512,
            ],
            key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
            psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
            ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
            support_sni=True,
        )

    def set_profile(self, profile):
        """Initializes the client according to the given profile.

        The following profiles are supported:

        - :obj:`tlsmate.tls.Profile.INTEROPERABILITY`

          - TLS Versions 1.0 - 1.3
          - ECDHE cipher & RSA-based key transport
          - AESGCM, AES, CHACHA_POLY and 3DES as last resort
          - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
          - Resumption, encrypt-then-mac, extended-master-secret
          - pskmode psk_dhe

        - :obj:`tlsmate.tls.Profile.LEGACY`

          - TLS Versions 1.0 - 1.2
          - ECDHE cipher & DHE & RSA-based key transport
          - AESGCM, AES, CHACHA_POLY and 3DES as last resort
          - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
          - Resumption, encrypt-then-mac, extended-master-secret

        - :obj:`tlsmate.tls.Profile.MODERN`

          - TLS Versions 1.2 + 1.3
          - ECDHE cipher
          - AESGCM, CHACHA_POLY
          - signatures: ECDSA + PSS_RSAE
          - Resumption, encrypt-then-mac, extended-master-secret
          - pskmode psk_dhe

        - :obj:`tlsmate.tls.Profile.TLS13`

          - TLS Version 1.3
          - AESGCM + CHACHA_POLY
          - pskmode psk_dhe

        Arguments:
            profile (:obj:`tlsmate.tls.Profile`): the profile to which the client
                shall be initialized.
        """
        if profile is tls.Profile.INTEROPERABILITY:
            self._set_profile_interoperability()
        elif profile is tls.Profile.TLS13:
            self._set_profile_tls13()
        elif profile is tls.Profile.MODERN:
            self._set_profile_modern()
        elif profile is tls.Profile.LEGACY:
            self._set_profile_legacy()
        else:
            raise ValueError(f"client profile {profile} unknown")

    def create_connection(self, host=None, port=None):
        """Create a new connection object

        Arguments:
            host (str): the host to contact. If not given, the host
                is taken from the configuration. The given string can be
                a URL or an IP address.
            port (int): the port number

        Returns:
            :obj:`tlsmate.connection.TlsConnection`: the created connection object
        """

        self._host = host if host is not None else self.config.get("host")

        interval = self.config.get("interval")
        if interval:
            time.sleep(interval / 1000)

        return TlsConnection(self._tlsmate, self._host)

    def save_session_state_id(self, session_state):
        """Save a session state

        Args:
            session_state (:obj:`tlsmate.structs.SessionStateId`): A session
                state to be stored on the client level, usable to resume
                connections using the session_id
        """
        self.session_state_id = session_state

    def get_session_state_id(self):
        """Get the session state (id)

        Returns:
            :obj:`tlsmate.structs.SessionStateId`: the session state to resume a
            session from
        """
        return self.session_state_id

    def save_session_state_ticket(self, session_state):
        """Save a session state

        Args:
            session_state (:obj:`tlsmate.structs.SessionStateId`): A session state to be
                stored on the client level, usable to resume connections using the
                session ticket.
        """
        self.session_state_ticket = session_state

    def get_session_state_ticket(self):
        """Get the session state (ticket)

        Returns:
            :obj:`tlsmate.structs.SessionStateTicket`: the session state to resume a
            session from
        """
        return self.session_state_ticket

    def save_psk(self, psk):
        """Save a TLS1.3 PSK

        Arguments:
            psk (:obj:`tlsmate.structs.Psk`): A pre-shared key be stored on the
                client level, usable to resume connections using the pre-shared key
                extension.
        """
        self.psks.append(psk)

    def get_sni(self):
        """Get the current SNI

        Returns:
            str: the SNI

        Raises:
            ValueError: if no SNI can be determined
        """

        sni = self.config.get("sni")
        if sni is not None:
            return sni

        sni = self._host
        if sni is not None:
            return sni

        return self.config.get("host")

    def client_hello(self):
        """Populate a ClientHello message according to the current client profile

        Returns:
            :obj:`tlsmate.msg.ClientHello`: the ClientHello object
        """
        msg = ClientHello()
        max_version = max(self.profile.versions)
        if max_version is tls.Version.TLS13:
            msg.version = tls.Version.TLS12

        else:
            msg.version = max_version

        msg.random = None  # will be provided autonomously

        if (
            self.profile.support_session_ticket
            and self.session_state_ticket is not None
        ):
            msg.session_id = bytes.fromhex("dead beef")

        elif self.profile.support_session_id and self.session_state_id is not None:
            msg.session_id = self.session_state_id.session_id

        else:
            msg.session_id = b""

        msg.cipher_suites = self.profile.cipher_suites[:]

        msg.compression_methods = self.profile.compression_methods
        if msg.version == tls.Version.SSL30:
            msg.extensions = None

        else:
            if self.profile.support_sni:
                msg.extensions.append(
                    ext.ExtServerNameIndication(host_name=self.get_sni())
                )

            if self.profile.support_extended_master_secret:
                msg.extensions.append(ext.ExtExtendedMasterSecret())

            if self.profile.ec_point_formats is not None:
                msg.extensions.append(
                    ext.ExtEcPointFormats(
                        ec_point_formats=self.profile.ec_point_formats
                    )
                )

            if self.profile.supported_groups is not None:
                if max_version is tls.Version.TLS13 or bool(
                    utils.filter_cipher_suites(
                        msg.cipher_suites, key_exch=[tls.KeyExchangeType.ECDH]
                    )
                ):
                    msg.extensions.append(
                        ext.ExtSupportedGroups(
                            supported_groups=self.profile.supported_groups
                        )
                    )

            if self.profile.support_status_request_v2 is not tls.StatusType.NONE:
                msg.extensions.append(
                    ext.ExtStatusRequestV2(
                        status_type=self.profile.support_status_request_v2
                    )
                )

            if self.profile.support_status_request:
                msg.extensions.append(ext.ExtStatusRequest())

            # RFC5246, 7.4.1.4.1.: Clients prior to TLS12 MUST NOT send this extension
            if (
                self.profile.signature_algorithms is not None
                and max_version >= tls.Version.TLS12
            ):
                msg.extensions.append(
                    ext.ExtSignatureAlgorithms(
                        signature_algorithms=self.profile.signature_algorithms
                    )
                )

            if self.profile.support_encrypt_then_mac:
                msg.extensions.append(ext.ExtEncryptThenMac())

            if self.profile.support_session_ticket:
                kwargs = {}
                if self.session_state_ticket is not None:
                    kwargs["ticket"] = self.session_state_ticket.ticket

                msg.extensions.append(ext.ExtSessionTicket(**kwargs))

            if self.profile.heartbeat_mode:
                msg.extensions.append(
                    ext.ExtHeartbeat(heartbeat_mode=self.profile.heartbeat_mode)
                )

            if tls.Version.TLS13 in self.profile.versions:
                if self._tlsmate.recorder.inject(
                    client_auth=self._tlsmate.client_auth.supported()
                ):
                    msg.extensions.append(ext.ExtPostHandshakeAuth())

                self._key_share_objects = []

                msg.extensions.append(
                    ext.ExtSupportedVersions(
                        versions=sorted(self.profile.versions, reverse=True)
                    )
                )
                # TLS13 key shares: enforce the same sequence as in supported groups
                if self.profile.key_shares:
                    key_shares = [
                        group
                        for group in self.profile.supported_groups
                        if group in self.profile.key_shares
                    ]
                    msg.extensions.append(ext.ExtKeyShare(key_shares=key_shares))

                if self.profile.early_data is not None:
                    msg.extensions.append(ext.ExtEarlyData())

                if self.profile.support_psk and self.psks:
                    msg.extensions.append(
                        ext.ExtPskKeyExchangeMode(
                            modes=self.profile.psk_key_exchange_modes
                        )
                    )
                    msg.extensions.append(ext.ExtPreSharedKey(psks=self.psks[:1]))

        return msg



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.heartbleed import ScanHeartbleed
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary


class TestHeartbleed(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_heartbeat_openssl1_0_2"
    recorder_yaml = "recorder_heartbleed"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa --no-cert-chain "
        "--ca-file ca-certificates "
        "-- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_1e

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        ScanHeartbleed(tlsmate).run()
        profile = server_profile.make_serializable()
        assert profile["vulnerabilities"]["heartbleed"] == "VULNERABLE"


if __name__ == "__main__":
    TestHeartbleed().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    path = pathlib.Path(__file__)
    name = "Heartbeat"
    cipher_suite = tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_2

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12

    def run(self, tlsmate, is_replaying=False):
        """The basic scenario to be recorded or replayed.
        """
        client = tlsmate.client
        client.init_profile()

        client.profile.versions = [self.version]
        client.profile.cipher_suites = [self.cipher_suite]
        client.profile.supported_groups = self.supported_groups
        client.profile.signature_algorithms = self.signature_algorithms

        end_of_tc_reached = False
        client.profile.heartbeat_mode = tls.HeartbeatMode.PEER_ALLOWED_TO_SEND
        with client.create_connection() as conn:
            conn.handshake()
            self.server_input("B\n", timeout=200)
            conn.timeout(500)
            request = msg.HeartbeatRequest()
            request.payload = b"abracadabra"
            request.payload_length = len(request.payload)
            request.padding = b"\xff" * 16
            conn.send(request)
            response = conn.wait(msg.HeartbeatResponse)
            assert response.payload_length == request.payload_length
            assert response.payload == request.payload
            end_of_tc_reached = True
        assert end_of_tc_reached is True
        return conn


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module containing exception definitions
"""


class TlsmateException(Exception):
    """A class all exception for tlsmate are based on.
    """


class ServerMalfunction(TlsmateException):
    """Exception raised in case the server response contains unrecoverable errors.

    This exception basically indicates a specification violation by the server.

    Attributes:
        issue (:obj:`tlsmate.tls.ServerIssue`): the reason for the exception
        message (:obj:`tlsmate.tls.HandshakeType`): the message, if applicable
        extension (:obj:`tlsmate.tls.Extension`): the extension, if applicable
    """

    def __init__(self, issue, message=None, extension=None):
        super().__init__(issue.value)
        self.issue = issue
        self.message = message
        self.extension = extension


class TlsConnectionClosedError(TlsmateException):
    """Exception raised when the TLS connection is closed unexpectedly.

    Attributes:
        exc(Exception): the original exception
    """

    def __init__(self, exc=None):
        self.exc = exc


class TlsMsgTimeoutError(TlsmateException):
    """Exception raised when a message is not received within a given timeout.
    """

    pass


class CurveNotSupportedError(TlsmateException):
    """Exception raised when a curve is negotiated which is not supported.

    Attributes:
        message (str): A human readable string providing the cause
        curve (:class:`tlsmate.tls.SupportedGroups`): The curve has been
            offered by the client, and selected by the server, but it is not
            supported for a full key exchange.
    """

    def __init__(self, message, curve):
        self.message = message
        self.curve = curve


class ScanError(TlsmateException):
    """Exception which might occur during a scan.

    The exception will be raised if an abnormal condition during a scan is
    detected.

    Attributes:
        message (str): A human readable string describing the cause.
    """

    def __init__(self, message):
        self.message = message


class OcspError(TlsmateException):
    """Exception for OCSP errors

    Attributes:
        issue (str): A human readable string describing the cause.
    """

    def __init__(self, issue):
        self.issue = issue


class UntrustedCertificate(TlsmateException):
    """Exception for unsuccessful certificate (chain) validation.

    Attributes:
        issue (str): A human readable string describing the cause.
    """

    def __init__(self, issue):
        self.issue = issue


class ServerParmsSignatureInvalid(TlsmateException):
    """More user friendly exception than cryptography.exception.InvalidSignature
    """



# -*- coding: utf-8 -*-
"""Module containing a classes to abstract the socket
"""
# import basic stuff
import socket
import select
import logging
import time
import sys

# import own stuff
from tlsmate import utils
from tlsmate import tls
from tlsmate.exception import TlsConnectionClosedError, TlsMsgTimeoutError
from tlsmate import recorder
from tlsmate import resolver

# import other stuff


class Socket(object):
    """Class implementing the socket interface.

    Arguments:
        server (str): The name of server to connect to.
        port (int): The port number to connect to.
        recorder (:obj:`tlsmate.recorder.Recorder`): The recorder object
    """

    def __init__(self, tlsmate):
        self._socket = None
        self._config = tlsmate.config
        self._recorder = tlsmate.recorder
        self._fragment_max_size = 16384

    def open_socket(self, l4_addr):
        """Opens a socket.

        Arguments:
            l4_addr (:obj:`tlsmate.structs.TransportEndpoint`): The L4-endpoint,
                consisting of the IP-address and the port.
        """

        if l4_addr.host_type is tls.HostType.HOST:
            l4_addr = resolver.get_ip_endpoint(l4_addr)

        if l4_addr.host_type is tls.HostType.IPV4:
            family = socket.AF_INET
            addr_info = (l4_addr.host, l4_addr.port)

        else:
            family = socket.AF_INET6
            addr_info = (l4_addr.host, l4_addr.port, 0, 0)

        if self._recorder.is_injecting():
            return

        try:
            self._socket = socket.socket(family, socket.SOCK_STREAM)
            self._socket.settimeout(5.0)
            self._socket.connect(addr_info)
            self._socket.settimeout(None)
            addr = self._socket.getsockname()
            laddr = addr[0]
            lport = addr[1]
            addr = self._socket.getpeername()
            raddr = addr[0]
            rport = addr[1]

        except OSError:
            utils.exit_with_error(f"Cannot open TCP connection to {l4_addr}")

        if self._config.get("progress"):
            sys.stderr.write(".")
            sys.stderr.flush()

        logging.info(f"{utils.Log.time()}: Socket opened")
        logging.info(f"local address: {laddr}:{lport}")
        logging.info(f"remote address: {raddr}:{rport}")

    def close_socket(self):
        """Closes a socket.
        """

        if self._socket is not None:
            logging.info(f"{utils.Log.time()}: closing socket")
            self._socket.close()
            self._socket = None

    def sendall(self, data):
        """Sends data to the network.

        Arguments:
            data (bytes): The data to send.
        """

        cont = self._recorder.trace_socket_sendall(data)
        if cont:
            if self._socket is None:
                return

            self._socket.sendall(data)

    def recv_data(self, timeout=5):
        """Wait for data from the network.

        Arguments:
            timeout (float): The maximum time to wait in seconds.

        Returns:
            bytes: The bytes received or None if the timeout expired.

        Raises:
            TlsMsgTimeoutError: If no data is received within the given timeout
            TlsConnectionClosedError: If the connection was closed.
        """

        data = self._recorder.inject_socket_recv()
        if data is None:
            if self._socket is None:
                self.open_socket()

            start = time.time()
            rfds, wfds, efds = select.select([self._socket], [], [], timeout)
            timeout = time.time() - start
            if not rfds:
                self._recorder.trace_socket_recv(timeout, recorder.SocketEvent.TIMEOUT)
                raise TlsMsgTimeoutError

            try:
                data = self._socket.recv(self._fragment_max_size)

            except ConnectionResetError as exc:
                self._recorder.trace_socket_recv(timeout, recorder.SocketEvent.CLOSURE)
                self.close_socket()
                raise TlsConnectionClosedError(exc)

            self._recorder.trace_socket_recv(
                timeout, recorder.SocketEvent.DATA, data=data
            )
        if data == b"":
            self.close_socket()
            raise TlsConnectionClosedError()

        return data



# -*- coding: utf-8 -*-
"""Module defining some utilities
"""
# import basic stuff
import time
import logging
import json
import os
import sys
import argparse

# import own stuff
from tlsmate import tls
from tlsmate import mappings
from tlsmate import structs
from tlsmate import pdu

# import other stuff
import yaml
import pathlib


class BooleanOptionalAction(argparse.Action):
    """Class to support --flag and --no-flag arguments.

    Will be natively supported by Python3.9
    """

    def __init__(
        self,
        option_strings,
        dest,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):

        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        if help is not None and default is not None:
            help += f" (default: {default})"

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self):
        return " | ".join(self.option_strings)


def serialize_data(data, file_name=None, replace=True, use_json=False, indent=4):
    """Serialize the data object to JSON or yaml.

    Arguments:
        data: the serializable data structure
        file_name (str or :obj:`pathlib.Path`): the file to write the serialized
            data to. If not given, the serialized object is printed on STDOUT.
        replace (bool): If True, allow overwriting an existing file. Defaults to True.
            This argument is only evaluated if a file_name is given.
        json (bool): If True, use JSON, else use Yaml.
        indent (int): The indentation to apply.
    """

    if file_name is not None:
        if not replace and pathlib.Path(file_name).exists():
            form = "JSON" if use_json else "Yaml"
            print(f"File {file_name} existing. {form}-file not generated")
            return

        with open(file_name, "w") as fd:
            if use_json:
                json.dump(data, fd, indent=indent, sort_keys=True)

            else:
                yaml.dump(data, fd, indent=indent)
    else:
        if use_json:
            print(json.dumps(data, indent=indent, sort_keys=True))

        else:
            print(yaml.dump(data, indent=indent))


def deserialize_data(file_name):
    """Deserialize from a JSON- or Yaml-file.

    Arguments:
        file_name (str or :obj:`pathlib.Path`): the full file name

    Returns:
        object: the deserialized object
    """

    with open(file_name) as fd:
        return yaml.safe_load(fd)


def fold_string(text, max_length, sep=" "):
    """Splits a string into lines of the given length.

    Arguments:
        text (str): the string to split
        length (int): the maximum length of the line
        sep (str): the separator where splitting the text is allowed

    Returns:
        list: the list of strings
    """

    ret_lines = []
    tokens = []
    length = 0
    sep_len = len(sep)

    for token in text.split(sep):
        token_len = len(token)
        if length + token_len + sep_len > max_length:
            if tokens:
                ret_lines.append(sep.join(tokens) + sep)
                tokens = [token]
                length = token_len

            else:
                ret_lines.append(token + sep)
                tokens = []
                length = 0

        else:
            tokens.append(token)
            length += token_len + sep_len

    if tokens:
        ret_lines.append(sep.join(tokens))

    return ret_lines


def get_cipher_suite_details(cipher_suite):
    """Get details for a given cipher suite

    Arguments:
        cipher_suite (:class:`tlsmate.tls.CipherSuite`): The given cipher suite.

    Returns:
        :obj:`tlsmate.structs.CipherSuiteDetails`:
        The structure with the detailed info regarding the given cipher suite.
    """

    cs = mappings.supported_cipher_suites.get(cipher_suite)
    if cs is None:
        return None

    ciph = mappings.supported_ciphers[cs.cipher]
    key = mappings.key_exchange.get(cs.key_ex)
    if key is not None:
        key_ex_supp = key.key_ex_supported

    else:
        key_ex_supp = False

    mac_struct = mappings.supported_macs.get(cs.mac)

    return structs.CipherSuiteDetails(
        cipher_suite=cipher_suite,
        full_hs=(key_ex_supp and ciph.cipher_supported),
        key_exchange_supported=key_ex_supp,
        key_algo=cs.key_ex,
        key_algo_struct=key,
        cipher=cs.cipher,
        cipher_struct=ciph,
        mac=cs.mac,
        mac_struct=mac_struct,
    )


def filter_cipher_suites(
    cs_list,
    key_exch=None,
    key_auth=None,
    key_algo=None,
    cipher_type=None,
    cipher_prim=None,
    cipher=None,
    cipher_mode=None,
    mac=None,
    version=None,
    full_hs=None,
    key_exchange_supported=None,
    remove=False,
):
    """Filters a list of cipher suites.

    Various match conditions can be specified. A cipher suite is filtered, if all
    given conditions match (AND-logic). All filtered cipher suites are returned in
    a list.

    Arguments:
        cs_list (list of :class:`tlsmate.tls.CipherSuite`): A list of cipher
            suites to be filtered.
        key_algo (list of :class:`tlsmate.tls.KeyExchangeAlgorithm`): Optional
            match condition. If the key_algo (a combination of key_exch and key_auth,
            e.g. "DHE_RSA") of a cipher suite is in the given list, it is a match.
        key_exch (list of :class:`tlsmate.tls.KeyExchangeType`): Optional
            match condition. If the key_exch (e.g. "ECDH") of a cipher suite is in
            the given list, it is a match.
        key_auth (list of :class:`tlsmate.tls.KeyAuthentication`): Optional
            match condition. If the key_auth (e.g. "ECDSA") of a cipher suite is in
            the given list, it is a match.
        cipher_type (list of :class:`tlsmate.tls.CipherType`): Optional
            match condition. If the cipher_type (e.g. "STREAM") of a cipher suite
            is in the list, it is a match.
        cipher_prim (list of :class:`tlsmate.tls.CipherPrimitive`): Optional
            match condition. If the cipher_prim (e.g. "AES") of a cipher suite is in
            the given list, it is a match.
        cipher (list of :class:`tlsmate.tls.SymmetricCipher`): Optional
            match condition. If the cipher (e.g. "AES_256_CCM_8") of a cipher suite
            is in the given list, it is a match.
        mac (list of :class:`tlsmate.tls.HashPrimitive`): Optional match
            condition. If the mac (e.g. "SHA384") of a cipher suite is in the give
            list, it is a match.
        version (:class:`tlsmate.tls.Version`): Optional match condition.
            Cipher suites supported by the given TLS version are a match. This is
            rather rudimentary implemented: AEAD ciphers only for TLS1.2, specific
            ciphers only for TLS1.3.
        full_hs (bool): Optional match condition. If the implementation supports a
            full handshake with a cipher suite, i.e. an encrypted connection can
            successfully established, it is a match. It means the key exchange,
            the symmetric cipher and the hash primitive are all supported.
        key_exchange_supported (bool): Optional match condition. If the key exchange
            for a cipher suite is supported, it is a match. Note, that this does not
            mean that the symmetric cipher is supported as well.
        remove (bool): An indication if the filtered cipher suites shall be removed
            from the original list of cipher suites. Defaults to False.

    Returns:
        list of :class:`tlsmate.tls.CipherSuite`:
        The list of cipher suites that match all the given conditions.
    """

    filter_funcs = []
    if key_algo is not None:
        filter_funcs.append(lambda cs: cs.key_algo in key_algo)

    if key_exch is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.key_algo_struct, "key_ex_type", None) in key_exch
        )

    if key_auth is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.key_algo_struct, "key_auth", None) in key_auth
        )

    if cipher_type is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.cipher_struct, "c_type", None) in cipher_type
        )

    if cipher_prim is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.cipher_struct, "primitive", None) in cipher_prim
        )

    if cipher is not None:
        filter_funcs.append(lambda cs: cs.cipher in cipher)

    if mac is not None:
        filter_funcs.append(lambda cs: cs.mac in mac)

    if version in [tls.Version.SSL30, tls.Version.TLS10, tls.Version.TLS11]:
        filter_funcs.append(
            lambda cs: cs.key_algo is not tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
            and (
                cs.cipher_struct is None
                or cs.cipher_struct.c_type is not tls.CipherType.AEAD
            )
        )

    if version is tls.Version.TLS12:
        filter_funcs.append(
            lambda cs: cs.key_algo is not tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
        )

    if version is tls.Version.TLS13:
        filter_funcs.append(
            lambda cs: cs.key_algo is tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
        )

    if full_hs is not None:
        filter_funcs.append(lambda cs: cs.full_hs is full_hs)

    if key_exchange_supported is not None:
        filter_funcs.append(
            lambda cs: cs.key_exchange_supported is key_exchange_supported
        )

    if tls.CipherSuite.TLS_FALLBACK_SCSV in cs_list:
        cs_list.remove(tls.CipherSuite.TLS_FALLBACK_SCSV)

    if tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV in cs_list:
        cs_list.remove(tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV)

    filtered = []
    for cs in cs_list:
        cs_details = get_cipher_suite_details(cs)
        if cs_details is None:
            continue

        match = True
        for filt_func in filter_funcs:
            if not filt_func(cs_details):
                match = False
                break

        if match:
            filtered.append(cs)

    if remove:
        for cs in filtered:
            cs_list.remove(cs)

    return filtered


def int_to_bytes(number):
    """Convert an integer to a byte string

    Arguments:
        number (int): the integer to convert

    Returns:
        bytes: just as many bytes as needed to represent the integer as an octet string
    """

    if number:
        return number.to_bytes((number.bit_length() + 7) // 8, "big")

    else:
        return b"\0"


def set_logging_format():
    """Initializes the format of log messages
    """

    logging.basicConfig(format="%(levelname)s: %(message)s")


def set_logging_level(level):
    """Sets the logging level

    Arguments:
        level (str): The logging level to use.
    """

    logging.getLogger().setLevel(level.upper())


class Log(object):
    """A class which implements relative time stamps.
    """

    start_time = None

    @classmethod
    def time(cls):
        """returns a time stamp relative to the time of the first call to this method.

        Returns:
            str: the time in seconds with 3 positions after the decimal point.
        """

        timestamp = time.time()
        if cls.start_time is None:
            cls.start_time = timestamp

        diff = timestamp - cls.start_time
        return f"Timestamp {diff:.3f}"


class Table(object):
    """Helper class for printing text in a table

    Attributes:
        indent (int): the indentation level, i.e. how many blanks shall be added
            on the left side of the table
        sep (str): the separator to print between adjacent columns within a row
    """

    def __init__(self, indent=0, sep=": "):
        self._indent = indent
        self._sep = sep
        self._nbr_columns = 0
        self._rows = []

    def row(self, *args):
        """Register a row

        Arguments:
            args (str): a complete row, one string for each column
        """
        self._nbr_columns = max(self._nbr_columns, len(args))
        cols = [col if type(col) is tuple else (col, len(col)) for col in args]
        self._rows.append(cols)

    def dump(self):
        """Print the table
        """

        if not self._nbr_columns:
            return

        cols = [1] * self._nbr_columns
        for row in self._rows:
            for idx, col in enumerate(row):
                cols[idx] = max(cols[idx], col[1])

        cols[-1] = 1
        for row in self._rows:
            print(" " * self._indent, end="")
            print(
                self._sep.join(
                    [
                        f"{col[0]:{cols[idx] + len(col[0]) - col[1]}}"
                        for idx, col in enumerate(row)
                    ]
                )
            )


def get_random_value():
    """Get a value suitable for a ClientHello or ServerHello

    Returns:
        bytes: 32 bytes of almost random data
    """

    random = bytearray()
    random.extend(pdu.pack_uint32(int(time.time())))
    random.extend(os.urandom(28))
    return random


def log_extensions(extensions):
    """Log extensions

    Arguments:
        extensions: the list of extensions to iterate over
    """

    for extension in extensions:
        extension = extension.extension_id
        logging.debug(f"extension {extension.value} {extension}")


def exit_with_error(error):
    """Abort tlsmate with an error message

    Arguments:
        error (str): the error message to print on stderr
    """

    sys.stderr.write(f"Error: {error}\n")
    sys.stderr.flush()
    sys.exit(1)



# -*- coding: utf-8 -*-
"""Module for host name handling
"""
# import basic stuff
import socket
import logging

# import own stuff
from tlsmate.exception import ScanError
from tlsmate import tls
from tlsmate import structs
from tlsmate import utils

# import other stuff


def determine_l4_addr(host, port):
    """Determine type of the host

    Arguments:
        host (str): the given host. This might be an IP address or a hostname.
        port (int): the given port.

    Returns:
        :obj:`tlsmate.structs.TransportEndpoint`: The structure representing a transport
        protocol endpoint.
    """

    try:
        socket.inet_pton(socket.AF_INET, host)
        host_type = tls.HostType.IPV4

    except OSError:
        try:
            socket.inet_pton(socket.AF_INET6, host)
            host_type = tls.HostType.IPV6

        except OSError:
            host_type = tls.HostType.HOST

    return structs.TransportEndpoint(host=host, port=port, host_type=host_type)


_resolved = {}


def resolve_hostname(host_name):
    """Resolve a hostname into sets of IPv4 and IPv6 addresses

    Arguments:
        host_name (str): the host name to resolve. May not be an IP address.

    Returns:
        :obj:`tlsmate.structs.ResolvedHost`: the structure for a resolved host name.
    """

    if host_name not in _resolved:
        try:
            logging.debug(f"Performing DNS lookup for {host_name}")
            ips = socket.getaddrinfo(host_name, None, type=socket.SOCK_STREAM)

        except socket.gaierror:
            utils.exit_with_error(f"Cannot resolve domain name {host_name}")

        ipv4_addresses = [item[4][0] for item in ips if item[0] is socket.AF_INET]
        ipv6_addresses = [item[4][0] for item in ips if item[0] is socket.AF_INET6]
        _resolved[host_name] = structs.ResolvedHost(
            ipv4_addresses=ipv4_addresses, ipv6_addresses=ipv6_addresses
        )
        for ipv4 in ipv4_addresses:
            logging.debug(f"IPv4 address: {ipv4}")

        for ipv6 in ipv6_addresses:
            logging.debug(f"IPv6 address: {ipv6}")

    return _resolved[host_name]


def get_ip_endpoint(l4_addr):
    """Resolve the hostname, if applicable.

    Arguments:
        l4_addr (:obj:`tlsmate.structs.TransportEndpoint`): the l4_addr to resolve

    Returns:
        :obj:`tlsmate.structs.TransportEndpoint`: the l4_addr, either with an IPv4
        or IPv6 address.
    """

    if l4_addr.host_type is not tls.HostType.HOST:
        return l4_addr

    ips = resolve_hostname(l4_addr.host)
    if ips.ipv4_addresses:
        host = ips.ipv4_addresses[0]
        host_type = tls.HostType.IPV4

    else:
        if ips.ipv6_addresses:
            host = ips.ipv4_addresses[0]
            host_type = tls.HostType.IPV4

        else:
            raise ScanError(f"No IP address available for {l4_addr.host}")

    return structs.TransportEndpoint(host=host, port=l4_addr.port, host_type=host_type)



# -*- coding: utf-8 -*-
"""Implements a class to test the grease worker.
"""
import pathlib
from tlsmate.workers.ephemeral_key_reuse import ScanEphemeralKeyReuse
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_sig_algos_openssl1_0_2"
    recorder_yaml = "recorder_ephemeral_key_reuse"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_2

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        ScanEphemeralKeyReuse(tlsmate).run()
        profile = server_profile.make_serializable()
        ekr = profile["features"]["ephemeral_key_reuse"]
        assert ekr["tls12_dhe_reuse"] == "FALSE"
        assert ekr["tls12_ecdhe_reuse"] == "TRUE"
        assert ekr["tls13_dhe_reuse"] == "NA"
        assert ekr["tls13_ecdhe_reuse"] == "NA"


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)
